@@ -1,4 +1,4 @@
-from typing import Optional, List, Dict
+from typing import Optional, List, Dict, Any
 import logging
 import time
 import asyncio
@@ -6,45 +6,29 @@ import websockets
 import json
 import urllib
 import threading
+import re
 from concurrent.futures._base import CancelledError
 from dataclasses import asdict
 from aiohttp import web
-from rpc_gateway import messages, errors
+from rpc_gateway import errors, messages_pb2, utils, http_gateway
 from rpc_gateway.websocket_connection import WebsocketConnection, next_message_id
 from rpc_gateway.utils import await_sync
 
 logger = logging.getLogger(__name__)
 
 
-class GatewayConnection:
-    def __init__(self, gateway: 'Gateway', connection: websockets.WebSocketServerProtocol):
-        self.gateway = gateway
-        self.connection = connection
-        self.message_pump = WebsocketConnection(connection, lambda message: gateway.handle_request(self, message))
-        self.instances: List[str] = []
-
-    async def start(self):
-        await self.message_pump.start()
-
-    async def stop(self):
-        await self.message_pump.stop()
-
-    def register_instances(self, instances):
-        self.instances = list(set(*self.instances, *instances))
-
-
-class Gateway:
-    SERVER_MESSAGES = (messages.Method.GET, messages.Method.SET, messages.Method.CALL, messages.Method.LOCK,
-                       messages.Method.UNLOCK, messages.Method.METADATA)
+class Gateway(http_gateway.MessageHandler):
+    SERVER_MESSAGES = (messages_pb2.Method.GET, messages_pb2.Method.SET, messages_pb2.Method.CALL, messages_pb2.Method.LOCK,
+                       messages_pb2.Method.UNLOCK, messages_pb2.Method.METADATA)
 
     def __init__(self, host: str = 'localhost', port: int = 8888, http_port: int = 8887, auth_key: str = 'DEFAULT_KEY', event_loop: Optional[asyncio.AbstractEventLoop] = None):
         self.host = host
         self.port = port
-        self.http_port = http_port
         self.auth_key = auth_key
         self.logger = logger.getChild(self.__class__.__name__)
         self.websocket_connections: List[WebsocketConnection] = []
         self.websocket: Optional[websockets.WebSocketServer] = None
+        self.http_gateway = http_gateway.HttpGateway(self, host, http_port, auth_key)
         self.event_loop = event_loop or asyncio.get_event_loop()
         self.instances: Dict[str, WebsocketConnection] = {}
         self.instance_groups: Dict[str, List[str]] = {}
@@ -52,10 +36,6 @@ class Gateway:
         self.check_connections_task: Optional[asyncio.Task] = None
         self.running = False
 
-        self.http_server: Optional[web.Server] = None
-        self.http_runner: Optional[web.ServerRunner] = None
-        self.http_site: Optional[web.TCPSite] = None
-
     def start(self, wait = True):
         self.event_loop.run_until_complete(self._start(wait=False))
 
@@ -65,13 +45,7 @@ class Gateway:
     async def _start(self, wait = True):
         self.logger.info(f'Starting on ws://{self.host}:{self.port}')
         self.websocket = await websockets.serve(self.on_connection, self.host, self.port)
-
-        self.logger.info(f'Starting HTTP server on http://{self.host}:{self.http_port}, auth_key={self.auth_key}')
-        self.http_server = web.Server(self.handle_http_request)
-        self.http_runner = web.ServerRunner(self.http_server)
-        await self.http_runner.setup()
-        self.http_site = web.TCPSite(self.http_runner, self.host, self.http_port)
-        await self.http_site.start()
+        await self.http_gateway.start()
         self.running = True
         self.check_connections_task = self.check_connections()
 
@@ -92,7 +66,7 @@ class Gateway:
 
     async def _stop(self):
         self.running = False
-        await self.http_site.stop()
+        await self.http_gateway.stop()
         try:
             await asyncio.gather(*[websocket_connection.stop() for websocket_connection in self.websocket_connections])
         except asyncio.exceptions.CancelledError:
@@ -114,9 +88,9 @@ class Gateway:
 
         await websocket_connection.start()
 
-    async def _unlock_instance(self, instance_name: str):
+    async def _unlock_instance(self, instance_name: str) -> messages_pb2.Message:
         server = self.instances[instance_name]
-        return await server.request(messages.Request(method=messages.Method.UNLOCK, data=instance_name), raise_error=False)
+        return await server.send_request(messages_pb2.Method.UNLOCK, instance=instance_name, raise_error=False)
 
     def is_registered(self, group: str, instance: str) -> bool:
         if group not in self.instance_groups:
@@ -136,55 +110,57 @@ class Gateway:
     # Request Handlers
     #
 
-    async def handle_forward_request(self, request: messages.Request) -> messages.Response:
+    async def handle_forward_request(self, request: messages_pb2.Message) -> messages_pb2.Message:
         self.logger.info(f'Forwarding request to server: {request}')
-        instance = request.data['instance']
+        instance_name = request.request.instance
 
-        if instance not in self.instances:
-            return messages.Response(status=messages.Status.ERROR, data=errors.InstanceNotFoundError(f'Instance not found: {instance}'))
+        if instance_name not in self.instances:
+            return utils.build_error_response(errors.InstanceNotFoundError(f'Instance not found: {instance_name}'))
 
-        server = self.instances[request.data['instance']]
+        server = self.instances[instance_name]
 
         response = await server.request(request, raise_error=False)
         self.logger.info(f'Forwarding response to client: {response}')
 
         return response
 
-    async def handle_available_request(self, request: messages.Request) -> messages.Response:
-        if request.data['instance'] not in self.instances:
-            return messages.Response(status=messages.Status.NOT_FOUND)
+    async def handle_available_request(self, request: messages_pb2.Message) -> messages_pb2.Message:
+        if request.request.instance not in self.instances:
+            return utils.build_response(status=messages_pb2.Status.NOT_FOUND)
 
-        server = self.instances[request.data['instance']]
+        server = self.instances[request.request.instance]
         return await server.request(request)
 
-    async def handle_list_request(self, request: messages.Request) -> messages.Response:
-        if 'group' not in request.data or request.data['group'] is None:
+    async def handle_list_request(self, request: messages_pb2.Message) -> messages_pb2.Message:
+        group = utils.decode_data(request.data, request.encoding)
+
+        if not group:
             instances = self.instance_groups
         else:
-            group = request.data['group']
-
             if group not in self.instance_groups:
-                return messages.Response(status=messages.Status.ERROR, data=errors.GroupNotFoundError(f'Group not found: {group}'))
+                return utils.build_error_response(errors.GroupNotFoundError(f'Group not found: {group}'))
 
             instances = {group: self.instance_groups[group]}
 
-        return messages.Response(data=instances)
+        return utils.build_response(instances)
 
-    async def handle_register_request(self, websocket_connection: WebsocketConnection, request: messages.Request) -> messages.Response:
-        self.logger.info(f'Registering instances: {request.data}')
+    async def handle_register_request(self, websocket_connection: WebsocketConnection, request: messages_pb2.Message) -> messages_pb2.Message:
+        instances = utils.decode_data(request.data, request.encoding)
+
+        self.logger.info(f'Registering instances: {instances}')
 
         # check to see if this instance
-        for instance_name, instance_group in request.data:
+        for instance_name, instance_group in instances:
             if instance_name is None:
-                return messages.Response(status=messages.Status.ERROR, data=errors.InvalidMessageError(f'Cannot register instance with name: {instance_name}'))
+                return utils.build_error_response(errors.InvalidMessageError(f'Cannot register instance with name: {instance_name}'))
 
             if instance_group is None:
-                return messages.Response(status=messages.Status.ERROR, data=errors.InvalidMessageError(f'Cannot register instance with group: {instance_group}'))
+                return utils.build_error_response(errors.InvalidMessageError(f'Cannot register instance with group: {instance_group}'))
 
             if self.is_registered(instance_group, instance_name):
-                return messages.Response(status=messages.Status.ERROR, data=errors.InstanceAlreadyRegisteredError(f'Instance "{instance_name}" already registered with "{instance_group}" group'))
+                return utils.build_error_response(errors.InstanceAlreadyRegisteredError(f'Instance "{instance_name}" already registered with "{instance_group}" group'))
 
-        for instance_name, instance_group in request.data:
+        for instance_name, instance_group in instances:
             if instance_group not in self.instance_groups:
                 self.instance_groups[instance_group] = []
 
@@ -192,15 +168,16 @@ class Gateway:
             self.instances[instance_name] = websocket_connection
             self.websocket_instances[websocket_connection].append(instance_name)
 
-        return messages.Response()
+        return utils.build_response()
 
-    async def handle_deregister_request(self, websocket_connection: WebsocketConnection, request: messages.Request) -> messages.Response:
-        self.logger.info(f'Deregistering instances: {request.data}')
+    async def handle_deregister_request(self, websocket_connection: WebsocketConnection, request: messages_pb2.Message) -> messages_pb2.Message:
+        instances = utils.decode_data(request.data, request.encoding)
+        self.logger.info(f'Deregistering instances: {instances}')
 
         # find instance to deregister
-        for instance_name in request.data:
+        for instance_name in instances:
             if instance_name not in self.instances:
-                return messages.Response(status=messages.Status.ERROR, data=errors.InstanceNotFoundError(f'Instance "{instance_name}" not found'))
+                return utils.build_error_response(errors.InstanceNotFoundError(f'Instance "{instance_name}" not found'))
 
             instance_connection = self.instances[instance_name]
             self.instances.pop(instance_name)
@@ -210,26 +187,28 @@ class Gateway:
                 if instance_name in group_instances:
                     group_instances.remove(instance_name)
 
-        return messages.Response()
+        return utils.build_response()
 
-    # this is called by the GatewayConnection MessagePump when a new request is received
-    async def handle_request(self, websocket_connection: WebsocketConnection, request: messages.Request) -> messages.Response:
-        if request.method in self.SERVER_MESSAGES:
-            return await self.handle_forward_request(request)
+    # this is called by the WebsocketConnection MessagePump when a new request is received
+    async def handle_request(self, websocket_connection: WebsocketConnection, message: messages_pb2.Message) -> messages_pb2.Message:
+        method = message.request.method
 
-        if request.method == messages.Method.AVAILABLE:
-            return await self.handle_available_request(request)
+        if method in self.SERVER_MESSAGES:
+            return await self.handle_forward_request(message)
 
-        if request.method == messages.Method.REGISTER:
-            return await self.handle_register_request(websocket_connection, request)
+        if method == messages_pb2.Method.AVAILABLE:
+            return await self.handle_available_request(message)
 
-        if request.method == messages.Method.DEREGISTER:
-            return await self.handle_deregister_request(websocket_connection, request)
+        if method == messages_pb2.Method.REGISTER:
+            return await self.handle_register_request(websocket_connection, message)
 
-        if request.method == messages.Method.LIST:
-            return await self.handle_list_request(request)
+        if method == messages_pb2.Method.DEREGISTER:
+            return await self.handle_deregister_request(websocket_connection, message)
 
-        return messages.Response(status=messages.Status.ERROR, data=errors.InvalidMethodError(f'Invalid method: {request.method}'))
+        if method == messages_pb2.Method.LIST:
+            return await self.handle_list_request(message)
+
+        return utils.build_error_response(errors.InvalidMethodError(f'Invalid method: {method}'))
 
     async def handle_close(self, websocket_connection: WebsocketConnection):
         # check if this is a server connection
@@ -238,7 +217,9 @@ class Gateway:
 
             # remove registered instances
             for instance_name in instances:
-                self.instances.pop(instance_name)
+                if instance_name in self.instances:
+                    self.instances.pop(instance_name)
+
                 groups_to_remove = []
 
                 for group, group_instances in self.instance_groups.items():
@@ -249,67 +230,19 @@ class Gateway:
                         groups_to_remove.append(group)
 
                 for group in groups_to_remove:
-                    self.instance_groups.pop(group)
+                    if group in self.instance_groups:
+                        self.instance_groups.pop(group)
 
             self.websocket_instances.pop(websocket_connection)
 
         # send an error response for any in-progress requests
         for request_id, response_queue in websocket_connection.receive_queues.items():
-            await response_queue.put(messages.Response(status=messages.Status.ERROR, data=errors.ServerConnectionLostError(f'Server connection lost')))
+            await response_queue.put(utils.build_error_response(errors.ServerConnectionLostError(f'Server connection lost')))
 
         self.websocket_connections.remove(websocket_connection)
 
         self.logger.info(f'Connection from {websocket_connection.connection.remote_address} closed')
 
-    def check_authentication(self, request: web.BaseRequest) -> bool:
-        if 'Authentication' in request.headers:
-            auth_header = request.headers['Authentication']
-
-            if auth_header.startswith('AuthKey '):
-                auth_type, auth_key = auth_header.split(' ', maxsplit=2)
-
-                return auth_key == self.auth_key
-
-        elif 'auth_key' in request.query:
-            return request.query['auth_key'] == self.auth_key
-
-        return False
-
-    async def handle_http_request(self, request: web.BaseRequest) -> web.StreamResponse:
-        if request.method != 'POST':
-            return web.HTTPMethodNotAllowed(request.method, ['POST'])
-
-        if not self.check_authentication(request):
-            return web.json_response({'status': 1, 'data': 'Authentication required'}, status=401)
-
-        message_raw = await request.read()
-
-        try:
-            self.logger.info(f'Handling request: {message_raw}')
-            message = json.loads(message_raw)
-            request = messages.Request(id=next_message_id(), method=message['method'], data=message['data'])
-
-            if request.method in self.SERVER_MESSAGES:
-                response = await self.handle_forward_request(request)
-            elif request.method == messages.Method.AVAILABLE:
-                response = await self.handle_available_request(request)
-            elif request.method == messages.Method.LIST:
-                response = await self.handle_list_request(request)
-            else:
-                response = messages.Response(status=messages.Status.ERROR, data=errors.InvalidMethodError(f'Invalid method: {request.method}'))
-
-        except (KeyError, json.decoder.JSONDecodeError):
-            response = messages.Response(status=messages.Status.ERROR, data=errors.InvalidMessageError(f'Invalid message: {message_raw}'))
-
-        if response.status == messages.Status.ERROR:
-            response.data = repr(response.data)
-
-        response_dict = asdict(response)
-        response_dict.pop('id')
-        response_dict.pop('encoding')
-
-        return web.json_response(response_dict)
-
 
 class GatewayClient:
     def __init__(self, gateway_url: str = 'ws://localhost:8888', auth_key: str = 'DEFAULT_KEY'):
@@ -373,14 +306,14 @@ class GatewayClient:
         await self._on_stop()
         await self.websocket_connection.stop()
 
-    async def _log_and_handle_request(self, websocket_connection: WebsocketConnection, request: messages.Request) -> messages.Response:
+    async def _log_and_handle_request(self, websocket_connection: WebsocketConnection, request: messages_pb2.Message) -> messages_pb2.Message:
         self.logger.debug(f'Request from {websocket_connection}: {request}')
         response = await self._handle_request(websocket_connection, request)
         self.logger.debug(f'Response to {websocket_connection}: {response}')
 
         return response
 
-    async def _handle_request(self, websocket_connection: WebsocketConnection, request: messages.Request) -> messages.Response:
+    async def _handle_request(self, websocket_connection: WebsocketConnection, request: messages_pb2.Message) -> messages_pb2.Message:
         pass
 
     async def _handle_close(self, websocket_connection: WebsocketConnection):
