@@ -6,7 +6,7 @@ import json
 from dataclasses import dataclass, asdict
 from inspect import isclass, signature, getdoc, Signature, Parameter, _empty
 from concurrent.futures import ThreadPoolExecutor
-from rpc_gateway import errors, messages, gateway
+from rpc_gateway import errors, gateway, messages_pb2, utils
 from rpc_gateway.utils import await_sync
 
 logger = logging.getLogger(__name__)
@@ -138,7 +138,7 @@ class Server(gateway.GatewayClient):
 
     async def _register_gateway_instances(self, instances: Dict[str, Instance]):
         instance_data = [(instance.name, instance.group) for name, instance in instances.items()]
-        await self.websocket_connection.send_request(messages.Method.REGISTER, instance_data)
+        await self.websocket_connection.send_request(messages_pb2.Method.REGISTER, data=instance_data)
 
     def register(self, name: str, instance: Any, group: Optional[str] = None) -> Any:
         self.logger.info(f'Registering instance: {name}')
@@ -158,7 +158,7 @@ class Server(gateway.GatewayClient):
 
     def deregister(self, name: str) -> Any:
         self.logger.info(f'Deregistering instance: {name}')
-        self.websocket_connection.send_request_sync(messages.Method.DEREGISTER, [name])
+        self.websocket_connection.send_request_sync(messages_pb2.Method.DEREGISTER, data=[name])
 
     def _get_instance(self, instance_name: Union[type, str]) -> Instance:
         try:
@@ -169,59 +169,58 @@ class Server(gateway.GatewayClient):
         except KeyError:
             raise errors.InstanceNotFoundError(f'Instance not found: {instance_name}')
 
-    def _available(self, instance_name: str) -> messages.Response:
+    def _available(self, instance_name: str) -> messages_pb2.Message:
         if instance_name not in self.instances:
-            return messages.Response(status=messages.Status.NOT_FOUND)
+            return utils.build_response(status=messages_pb2.Status.NOT_FOUND)
 
         if self.instances[instance_name].locked:
-            return messages.Response(status=messages.Status.LOCKED)
+            return utils.build_response(status=messages_pb2.Status.LOCKED)
 
-        return messages.Response()
+        return utils.build_response()
 
-    def _lock(self, instance_name: str) -> messages.Response:
+    def _lock(self, instance_name: str) -> messages_pb2.Message:
         instance = self._get_instance(instance_name)
         instance.locked = True
 
-        return messages.Response()
+        return utils.build_response()
 
-    def _unlock(self, instance_name: str) -> messages.Response:
+    def _unlock(self, instance_name: str) -> messages_pb2.Message:
         instance = self._get_instance(instance_name)
         instance.locked = False
 
-        return messages.Response()
+        return utils.build_response()
 
-    def _metadata(self, instance_name: str) -> messages.Response:
+    def _metadata(self, instance_name: str) -> messages_pb2.Message:
         instance = self._get_instance(instance_name)
         metadata = InstanceMetadata.from_instance(instance)
 
-        return messages.Response(data=asdict(metadata))
+        return utils.build_response(data=asdict(metadata))
 
-    def _get(self, instance_name: str, attribute_path: str) -> messages.Response:
+    def _get(self, instance_name: str, attribute_path: str) -> messages_pb2.Message:
         try:
             instance = self._get_instance(instance_name)
             data = get_attribute_path(instance.instance, attribute_path.split('.'))
 
             if callable(data):
-                return messages.Response(status=messages.Status.METHOD)
+                return utils.build_response(status=messages_pb2.Status.METHOD)
 
             try:
                 serialized_data = pickle.dumps(data)
+                return utils.build_response(serialized_data=serialized_data)
             except (pickle.PickleError, TypeError):
-                return messages.Response(status=messages.Status.PROXY)
-
-            return messages.Response(data=data)
+                return utils.build_response(status=messages_pb2.Status.PROXY)
         except Exception as err:
-            return messages.Response(status=messages.Status.ERROR, data=err)
+            return utils.build_error_response(err)
 
-    def _set(self, instance_name: str, attribute_path: str, value: Any) -> messages.Response:
+    def _set(self, instance_name: str, attribute_path: str, value: Any) -> messages_pb2.Message:
         try:
             instance = self._get_instance(instance_name)
             set_attribute_path(instance.instance, attribute_path.split('.'), value)
-            return messages.Response()
+            return utils.build_response()
         except Exception as err:
-            return messages.Response(status=messages.Status.ERROR, data=err)
+            return utils.build_error_response(err)
 
-    def _call(self, instance_name: str, attribute_path: str, args: Union[List[Any], Dict[str, Any]], kwargs: Dict[str, Any]) -> messages.Response:
+    def _call(self, instance_name: str, attribute_path: str, args: Union[List[Any], Dict[str, Any]], kwargs: Dict[str, Any]) -> messages_pb2.Message:
         try:
             instance = self._get_instance(instance_name)
             method = get_attribute_path(instance.instance, attribute_path.split('.'))
@@ -231,39 +230,42 @@ class Server(gateway.GatewayClient):
                 args = [args[str(i)] for i in range(len(args))]
 
             data = method(*args, **kwargs)
-            return messages.Response(data=data)
+            return utils.build_response(data=data)
         except Exception as err:
-            return messages.Response(status=messages.Status.ERROR, data=err)
+            return utils.build_error_response(err)
 
     async def _run(self, *args: Any) -> Any:
         return await self.event_loop.run_in_executor(self.executor, *args)
 
-    async def _handle_request(self, websocket_connection: websockets.WebSocketCommonProtocol, request: messages.Request) -> messages.Response:
+    async def _handle_request(self, websocket_connection: websockets.WebSocketCommonProtocol, message: messages_pb2.Message) -> messages_pb2.Message:
         try:
-            if request.method == messages.Method.GET:
-                return await self._run(self._get, request.data['instance'], request.data['attribute'])
+            request = message.request
+            data = utils.decode_data(message.data, message.encoding)
 
-            if request.method == messages.Method.SET:
-                return await self._run(self._set, request.data['instance'], request.data['attribute'], request.data['value'])
+            if request.method == messages_pb2.Method.GET:
+                return await self._run(self._get, request.instance, request.attribute)
 
-            if request.method == messages.Method.CALL:
-                return await self._run(self._call, request.data['instance'], request.data['attribute'], request.data['args'], request.data['kwargs'])
+            if request.method == messages_pb2.Method.SET:
+                return await self._run(self._set, request.instance, request.attribute, data)
 
-            if request.method == messages.Method.AVAILABLE:
-                return await self._run(self._available, request.data['instance'])
+            if request.method == messages_pb2.Method.CALL:
+                return await self._run(self._call, request.instance, request.attribute, data['args'], data['kwargs'])
 
-            if request.method == messages.Method.LOCK:
-                return await self._run(self._lock, request.data['instance'])
+            if request.method == messages_pb2.Method.AVAILABLE:
+                return await self._run(self._available, request.instance)
 
-            if request.method == messages.Method.UNLOCK:
-                return await self._run(self._unlock, request.data['instance'])
+            if request.method == messages_pb2.Method.LOCK:
+                return await self._run(self._lock, request.instance)
 
-            if request.method == messages.Method.METADATA:
-                return await self._run(self._metadata, request.data['instance'])
+            if request.method == messages_pb2.Method.UNLOCK:
+                return await self._run(self._unlock, request.instance)
+
+            if request.method == messages_pb2.Method.METADATA:
+                return await self._run(self._metadata, request.instance)
         except KeyError:
-            return messages.Response(status=messages.Status.ERROR, data=errors.InvalidMessageError(f'Invalid message'))
+            return utils.build_error_response(errors.InvalidMessageError(f'Invalid message'))
 
-        return messages.Response(status=messages.Status.ERROR, data=errors.InvalidMethodError(f'Invalid method: {request.method}'))
+        return utils.build_error_response(errors.InvalidMethodError(f'Invalid method: {request.method}'))
 
 
 if __name__ == '__main__':
