@@ -1,14 +1,15 @@
-from typing import Optional, Dict, Any, Callable, Awaitable
+from typing import Optional, Dict, Any, Callable, Awaitable, Union
 import asyncio
 import pickle
 import json
 import logging
 import websockets
 from concurrent import futures
-from rpc_gateway import messages, errors, utils
+from google.protobuf.message import EncodeError, DecodeError
+from rpc_gateway import errors, utils, messages_pb2
 
 logger = logging.getLogger(__name__)
-RequestHandlerType = Callable[['WebsocketConnection', messages.Request], Awaitable[messages.Response]]
+RequestHandlerType = Callable[['WebsocketConnection', messages_pb2.Message], Awaitable[messages_pb2.Message]]
 CloseHandlerType = Callable[['MessagePump'], Awaitable]
 MAX_MESSAGE_ID = 99999
 last_message_id = 0
@@ -75,71 +76,72 @@ class WebsocketConnection:
         except futures.CancelledError:
             pass
 
-    def send_message_sync(self, message: messages.Message):
+    def send_message_sync(self, message: messages_pb2.Message):
         utils.await_sync(self.send_message(message), self.event_loop)
 
-    async def send_message(self, message: messages.Message):
+    async def send_message(self, message: messages_pb2.Message):
         if self.connection is None:
             raise errors.NotConnectedError('Must be connected to send message: {message}')
 
         await self.send_queue.put(message)
 
-    def send_request_sync(self, method: str, data: Any = None) -> messages.Response:
-        return utils.await_sync(self.send_request(method, data), self.event_loop)
+    def send_request_sync(self, method: str, instance: Optional[str] = None, attribute: Optional[str] = None,
+                          data: Any = None, encoding: messages_pb2.Encoding = messages_pb2.Encoding.PICKLE) -> messages_pb2.Message:
+        return utils.await_sync(self.send_request(method, instance, attribute, data, encoding), self.event_loop)
 
-    async def send_request(self, method: str, data: Any = None) -> messages.Response:
-        return await self.request(messages.Request(method=method, data=data))
+    async def send_request(self, method: str, instance: Optional[str] = None, attribute: Optional[str] = None,
+                          data: Any = None, encoding: messages_pb2.Encoding = messages_pb2.Encoding.PICKLE, raise_error: bool = True) -> messages_pb2.Message:
+        encoded_data = utils.encode_data(data, encoding)
+        request = messages_pb2.Request(method=method, instance=instance, attribute=attribute)
+        message = messages_pb2.Message(data=encoded_data, encoding=encoding, request=request)
+        return await self.request(message, raise_error=raise_error)
 
-    def send_response_sync(self, id: int, data: Any = None, status: int = messages.Status.SUCCESS):
-        utils.await_sync(self.send_response(id, data, status), self.event_loop)
-
-    async def send_response(self, id: int, data: Any = None, status: int = messages.Status.SUCCESS):
-        response = messages.Response(id=id, status=status, data=data)
-        await self.send_message(response)
-
-    def send_error_response_sync(self, id: int, error: Exception):
-        utils.await_sync(self.send_error_response(id, error), self.event_loop)
-
-    async def send_error_response(self, id: int, error: Exception):
-        response = messages.Response(id=id, status=messages.Status.ERROR, data=error)
-        await self.send_message(response)
-
-    def request_sync(self, request: messages.Request, raise_error=True) -> messages.Response:
+    def request_sync(self, request: messages_pb2.Message, raise_error=True) -> messages_pb2.Message:
         return utils.await_sync(self.request(request, raise_error))
 
-    async def request(self, request: messages.Request, raise_error=True) -> messages.Response:
+    async def request(self, request: messages_pb2.Message, raise_error=True) -> messages_pb2.Message:
         request.id = next_message_id()
         queue = self.receive_queues[request.id] = asyncio.Queue()
         await self.send_message(request)
 
         try:
-            response: messages.Response = await queue.get()
+            response: messages_pb2.Message = await queue.get()
             self.receive_queues.pop(request.id)
         except asyncio.TimeoutError:
             self.receive_queues.pop(request.id)
             raise errors.RequestTimeoutError(f'Request timed out while waiting for response: {request}')
 
-        if response.status == messages.Status.ERROR and raise_error:
-            raise response.data
+        # raise an error if this is an error response
+        if response.response.status == messages_pb2.Status.ERROR and raise_error:
+            decoded_error = utils.decode_data(response.data, response.encoding)
+
+            if response.encoding == messages_pb2.Encoding.JSON:
+                raise Exception(decoded_error)
+
+            if response.encoding == messages_pb2.Encoding.PICKLE:
+                raise decoded_error
+
+            raise errors.InvalidEncodingError(f'Invalid response encoding: {response.encoding}')
 
         return response
 
-    async def _handle_request(self, request: messages.Request):
+    async def _handle_request(self, request: messages_pb2.Message):
         id = request.id
-        encoding = request.encoding
-        response = await self.request_handler(self, request)
+        response: messages_pb2.Message = await self.request_handler(self, request)
         response.id = id
-        response.encoding = encoding
         await self.send_message(response)
 
     async def _message_send(self):
         while self.running:
-            message: messages.Message = await self.send_queue.get()
+            message: messages_pb2.Message = await self.send_queue.get()
 
             try:
-                message_bytes = messages.encode(message)
-            except (pickle.PickleError, AttributeError):
-                message_bytes = messages.Response(id=message.id, status=messages.Status.ERROR, data=errors.SerializationError(f'Cannot serialize response data')).to_pickle()
+                message_bytes = message.SerializeToString()
+            except EncodeError:
+                error = errors.SerializationError(f'Cannot serialize response data')
+                message = messages_pb2.Message(id=message.id, data=utils.encode_error(error, message.encoding))
+                message.response.status = messages_pb2.Status.ERROR
+                message_bytes = message.SerializeToString()
 
             self.logger.info(f'[{self.id}] > {message}')
             await self.connection.send(message_bytes)
@@ -149,21 +151,22 @@ class WebsocketConnection:
             while self.running:
                 message_bytes = await self.connection.recv()
                 try:
-                    message = messages.decode(message_bytes)
-                except (pickle.UnpicklingError, TypeError):
+                    message = messages_pb2.Message()
+                    message.ParseFromString(message_bytes)
+                except DecodeError:
                     self.logger.error(f'Invalid message: {message_bytes}')
                     continue
 
                 self.logger.info(f'[{self.id}] < {message}')
 
-                if isinstance(message, messages.Response):  # response
+                if message.HasField('response'):
                     if message.id not in self.receive_queues:
                         raise errors.InvalidMessageIdError(f'No request found for response ID: {message.id}')
 
                     queue = self.receive_queues[message.id]
                     await queue.put(message)
 
-                elif isinstance(message, messages.Request):  # request
+                elif message.HasField('request'):  # request
                     if self.request_handler is not None:
                         asyncio.create_task(self._handle_request(message))
 
